(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global = global || self, global.TaskListsElement = factory());
}(this, function () { 'use strict';

  const sortHandlers = new WeakMap();
  let state = null;
  function isDragging() {
    return !!state;
  }
  function sortable(el, sortStarted, sortFinished) {
    sortHandlers.set(el, {
      sortStarted,
      sortFinished
    });
    el.addEventListener('dragstart', onDragStart);
    el.addEventListener('dragenter', onDragEnter);
    el.addEventListener('dragend', onDragEnd);
    el.addEventListener('drop', onDrop);
    el.addEventListener('dragover', onDragOver);
  } // Determine if item2 is before item1 in the tree.

  function isBefore(item1, item2) {
    if (item1.parentNode === item2.parentNode) {
      let node = item1;

      while (node) {
        if (node === item2) return true;
        node = node.previousElementSibling;
      }
    }

    return false;
  }

  function isSameContainer(item1, item2) {
    return item1.closest('task-lists') === item2.closest('task-lists');
  }

  function onDragStart(event) {
    // Ignore selected text dragging within list items.
    if (event.currentTarget !== event.target) return;
    const target = event.currentTarget;
    if (!(target instanceof Element)) return;
    const sourceList = target.closest('.contains-task-list');
    if (!sourceList) return;
    target.classList.add('is-ghost'); // Add data to drag operation (required for Firefox).

    if (event.dataTransfer) {
      event.dataTransfer.setData('text/plain', target.textContent.trim());
    }

    if (!target.parentElement) return;
    const siblings = Array.from(target.parentElement.children);
    const sourceIndex = siblings.indexOf(target);
    const handlers = sortHandlers.get(target);

    if (handlers) {
      handlers.sortStarted(sourceList);
    }

    state = {
      didDrop: false,
      dragging: target,
      dropzone: target,
      sourceList,
      sourceSibling: siblings[sourceIndex + 1] || null,
      sourceIndex
    };
  }

  function onDragEnter(event) {
    // Allow dragging task list item, not text, links, or files.
    if (!state) return;
    const dropzone = event.currentTarget;
    if (!(dropzone instanceof Element)) return;

    if (!isSameContainer(state.dragging, dropzone)) {
      event.stopPropagation();
      return;
    } // Necessary for dragging


    event.preventDefault();

    if (event.dataTransfer) {
      event.dataTransfer.dropEffect = 'move';
    } // Make sure we weren't already dragging in this task list item.


    if (state.dropzone === dropzone) return; // Open an empty space in the list as a drop target.

    state.dragging.classList.add('is-dragging');
    state.dropzone = dropzone;

    if (isBefore(state.dragging, dropzone)) {
      dropzone.before(state.dragging);
    } else {
      dropzone.after(state.dragging);
    }
  }

  function onDrop(event) {
    // Allow dragging task list item, not text, links, or files.
    if (!state) return;
    event.preventDefault();
    event.stopPropagation();
    const dropzone = event.currentTarget;
    if (!(dropzone instanceof Element)) return; // Notify dragend listener drag ended in a valid drop target.

    state.didDrop = true;
    if (!state.dragging.parentElement) return;
    let newIndex = Array.from(state.dragging.parentElement.children).indexOf(state.dragging);
    const currentList = dropzone.closest('.contains-task-list');
    if (!currentList) return; // Don't notify listener if the list hasn't changed.

    if (state.sourceIndex === newIndex && state.sourceList === currentList) return; // Increment insertion index when moving item down in its own list.

    if (state.sourceList === currentList && state.sourceIndex < newIndex) {
      newIndex++;
    }

    const src = {
      list: state.sourceList,
      index: state.sourceIndex
    };
    const dst = {
      list: currentList,
      index: newIndex
    };
    const handlers = sortHandlers.get(state.dragging);

    if (handlers) {
      handlers.sortFinished({
        src,
        dst
      });
    }
  }

  function onDragEnd() {
    if (!state) return;
    state.dragging.classList.remove('is-dragging');
    state.dragging.classList.remove('is-ghost'); // If the drag ended in a valid drop target, the drop event listener will
    // have already set this flag to true. If not, it'll still be false, and
    // we'll know the drag was canceled.

    if (!state.didDrop) {
      state.sourceList.insertBefore(state.dragging, state.sourceSibling);
    } // Reset for next drag.


    state = null;
  }

  function onDragOver(event) {
    // Allow dragging task list item, not text, links, or files.
    if (!state) return;
    const dropzone = event.currentTarget;
    if (!(dropzone instanceof Element)) return;

    if (!isSameContainer(state.dragging, dropzone)) {
      event.stopPropagation();
      return;
    } // Necessary for dragging.


    event.preventDefault();

    if (event.dataTransfer) {
      event.dataTransfer.dropEffect = 'move';
    }
  }

  const observers = new WeakMap();
  class TaskListsElement extends HTMLElement {
    constructor() {
      super();
      this.addEventListener('change', event => {
        const checkbox = event.target;
        if (!(checkbox instanceof HTMLInputElement)) return;
        if (!checkbox.classList.contains('task-list-item-checkbox')) return;
        this.dispatchEvent(new CustomEvent('task-lists-check', {
          bubbles: true,
          detail: {
            position: position(checkbox),
            checked: checkbox.checked
          }
        }));
      });
      observers.set(this, new MutationObserver(syncState.bind(null, this)));
    }

    connectedCallback() {
      const observer = observers.get(this);

      if (observer) {
        observer.observe(this, {
          childList: true,
          subtree: true
        });
      }

      syncState(this);
    }

    disconnectedCallback() {
      const observer = observers.get(this);

      if (observer) {
        observer.disconnect();
      }
    }

    get disabled() {
      return this.hasAttribute('disabled');
    }

    set disabled(value) {
      if (value) {
        this.setAttribute('disabled', '');
      } else {
        this.removeAttribute('disabled');
      }
    }

    get sortable() {
      return this.hasAttribute('sortable');
    }

    set sortable(value) {
      if (value) {
        this.setAttribute('sortable', '');
      } else {
        this.removeAttribute('sortable');
      }
    }

    static get observedAttributes() {
      return ['disabled'];
    }

    attributeChangedCallback(name, oldValue, newValue) {
      if (oldValue === newValue) return;

      switch (name) {
        case 'disabled':
          syncDisabled(this);
          break;
      }
    }

  }
  const handleTemplate = document.createElement('template');
  handleTemplate.innerHTML = `
  <span class="handle">
    <svg class="drag-handle" aria-hidden="true" width="16" height="15" version="1.1" viewBox="0 0 16 15">
      <path d="M12,4V5H4V4h8ZM4,8h8V7H4V8Zm0,3h8V10H4v1Z"></path>
    </svg>
  </span>`;
  const initialized = new WeakMap(); // Only top-level lists are draggable, and nested lists drag with their parent item.

  function initItem(el) {
    if (initialized.get(el)) return;
    initialized.set(el, true);
    const taskList = el.closest('task-lists');
    if (!(taskList instanceof TaskListsElement)) return; // Single item task lists are not draggable.

    if (taskList.querySelectorAll('.task-list-item').length <= 1) return;
    const fragment = handleTemplate.content.cloneNode(true);
    const handle = fragment.querySelector('.handle');
    el.prepend(fragment);
    if (!handle) throw new Error('handle not found');
    handle.addEventListener('mouseenter', onHandleMouseOver);
    handle.addEventListener('mouseleave', onHandleMouseOut);
    sortable(el, onSortStart, onSorted); // Drag operations don't remove :hover styles, so manage drag handle hover state.

    el.addEventListener('mouseenter', onListItemMouseOver);
    el.addEventListener('mouseleave', onListItemMouseOut);
  }

  function onListItemMouseOver(event) {
    const item = event.currentTarget;
    if (!(item instanceof Element)) return;
    const list = item.closest('task-lists');
    if (!(list instanceof TaskListsElement)) return;

    if (list.sortable && !list.disabled) {
      item.classList.add('hovered');
    }
  }

  function onListItemMouseOut(event) {
    const item = event.currentTarget;
    if (!(item instanceof Element)) return;
    item.classList.remove('hovered');
  } // Returns the list item position as a (list index, item index) tuple.


  function position(checkbox) {
    const list = taskList(checkbox);
    if (!list) throw new Error('.contains-task-list not found');
    const index = Array.from(list.children).indexOf(checkbox.closest('.task-list-item'));
    return [listIndex(list), index];
  } // Finds the list item's parent task list.


  function taskList(el) {
    const parent = el.parentElement;
    return parent ? parent.closest('.contains-task-list') : null;
  } // Is the task list this element belongs to a root list?


  function isRootTaskList(el) {
    return taskList(el) === rootTaskList(el);
  } // Finds the highest task list in this element's ancestor chain.


  function rootTaskList(node) {
    const list = taskList(node);
    return list ? rootTaskList(list) || list : null;
  }

  function syncState(list) {
    const items = list.querySelectorAll('.contains-task-list > .task-list-item');

    for (const el of items) {
      if (isRootTaskList(el)) {
        initItem(el);
      }
    }

    syncDisabled(list);
  }

  function syncDisabled(list) {
    for (const el of list.querySelectorAll('.task-list-item')) {
      el.classList.toggle('enabled', !list.disabled);
    }

    for (const el of list.querySelectorAll('.task-list-item-checkbox')) {
      if (el instanceof HTMLInputElement) {
        el.disabled = list.disabled;
      }
    }
  } // Given a top-level task list, return its index location in the container. All
  // lists in the container, not just task lists, are indexed to match the
  // server-side Markdown parser's indexing.


  function listIndex(list) {
    const container = list.closest('task-lists');
    if (!container) throw new Error('parent not found');
    return Array.from(container.querySelectorAll('ol, ul')).indexOf(list);
  }

  const originalLists = new WeakMap();

  function onSortStart(srcList) {
    const container = srcList.closest('task-lists');
    if (!container) throw new Error('parent not found');
    originalLists.set(container, Array.from(container.querySelectorAll('ol, ul')));
  }

  function onSorted(_ref) {
    let src = _ref.src,
        dst = _ref.dst;
    const container = src.list.closest('task-lists');
    if (!container) return;
    const lists = originalLists.get(container);
    if (!lists) return;
    originalLists.delete(container);
    container.dispatchEvent(new CustomEvent('task-lists-move', {
      bubbles: true,
      detail: {
        src: [lists.indexOf(src.list), src.index],
        dst: [lists.indexOf(dst.list), dst.index]
      }
    }));
  } // Enable list item drag when handle is hovered.


  function onHandleMouseOver(event) {
    const target = event.currentTarget;
    if (!(target instanceof Element)) return;
    const item = target.closest('.task-list-item');
    if (!item) return;
    const list = item.closest('task-lists');
    if (!(list instanceof TaskListsElement)) return;

    if (list.sortable && !list.disabled) {
      item.setAttribute('draggable', 'true');
    }
  } // Disable list item drag so text selection works.


  function onHandleMouseOut(event) {
    if (isDragging()) return;
    const target = event.currentTarget;
    if (!(target instanceof Element)) return;
    const item = target.closest('.task-list-item');
    if (!item) return;
    item.setAttribute('draggable', 'false');
  }

  if (!window.customElements.get('task-lists')) {
    window.TaskListsElement = TaskListsElement;
    window.customElements.define('task-lists', TaskListsElement);
  }

  return TaskListsElement;

}));
